using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using System.Security.AccessControl;

namespace Stopper {
	public class EyelidBehaviorEditor : MVRScript {

        string savePath = "Saves\\EyelidBehavior";
		string fileExt = "eyelidbehavior";

        private DAZMeshEyelidControl eyelidControl;
        private List<string> morphNames = new List<string>();
        private HashSet<DAZMorph> modifiedMorphs = new HashSet<DAZMorph>();

        private bool morphsMeedUpdate = false;
        private bool needRebuildMorphs = false;


        private float lookInwardTopEyelidFactor = 1f;
        private float lookInwardBottomEyelidFactor = 1f;
        private float lookOutwardTopEyelidFactor = 1f;
        private float lookOutwardBottomEyelidFactor = 1f;
        private DAZMorph LeftTopEyelidLeftMorph;
        private DAZMorph RightTopEyelidLeftMorph;
        private DAZMorph LeftBottomEyelidLeftMorph;
        private DAZMorph RightBottomEyelidLeftMorph;
        private DAZMorph LeftTopEyelidRightMorph;
        private DAZMorph RightTopEyelidRightMorph;
        private DAZMorph LeftBottomEyelidRightMorph;
        private DAZMorph RightBottomEyelidRightMorph;
        private DAZMorph dummyMorph = new DAZMorph();

        private bool leftRightSelected;

        public override void Init() {
			eyelidControl = containingAtom.GetComponentInChildren<DAZMeshEyelidControl>();
            //RefreshMorphList();

            // Preset code was adapted from AcidBubbles' Collider Editor
			var loadPresetUI = CreateButton("Load Preset");
			loadPresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var shortcuts = FileManagerSecure.GetShortCutsForDirectory(savePath);
				SuperController.singleton.GetMediaPathDialog(LoadPreset, fileExt, savePath, false, true, false, null, false, shortcuts);
			});

			var savePresetUI = CreateButton("Save Preset");
			savePresetUI.button.onClick.AddListener(() =>
			{
				FileManagerSecure.CreateDirectory(savePath);
				var fileBrowserUI = SuperController.singleton.fileBrowserUI;
				fileBrowserUI.SetTitle("Save Preset");
				fileBrowserUI.fileRemovePrefix = null;
				fileBrowserUI.hideExtension = false;
				fileBrowserUI.keepOpen = false;
				fileBrowserUI.fileFormat = fileExt;
				fileBrowserUI.defaultPath = savePath;
				fileBrowserUI.showDirs = true;
				fileBrowserUI.shortCuts = null;
				fileBrowserUI.browseVarFilesAsDirectories = false;
				fileBrowserUI.SetTextEntry(true);
				fileBrowserUI.Show(SavePreset);
				fileBrowserUI.fileEntryField.text = (int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds + "." + fileExt;
				fileBrowserUI.ActivateFileNameField();
			});

            // Blink Controls
            AddSlider("Blink Space Min", 1f, 0.001f, 10, (val) => {
                eyelidControl.blinkSpaceMin = val;
            }, true, "F2", true);

            AddSlider("Blink Space Max", 7f, 0.001f, 10, (val) => {
                eyelidControl.blinkSpaceMax = val;
            }, true, "F2", true);

            AddSlider("Blink Time Min", 0.1f, 0.001f, 1f, (val) => {
                eyelidControl.blinkTimeMin = val;
            }, true, "F3", true);

            AddSlider("Blink Time Max", 0.4f, 0.001f, 1f, (val) => {
                eyelidControl.blinkTimeMin = val;
            }, true, "F3", true);

            AddSlider("Blink Up/Down Ratio", 0.4f, 0.00f, 1f, (val) => {
                eyelidControl.blinkDownUpRatio = val;
            }, true, "F3", false);

            AddSlider("Blink Bottom Eyelid Factor", 0.5f, 0.00f, 1f, (val) => {
                eyelidControl.blinkBottomEyelidFactor = val;
            }, true, "F3", false);

            // Look controls

            // Up / Down
            AddLabel("Spacer", "");
            AddLabel("UpDownControls", "Up And Down");
            
            AddSlider("Look Up Top Eyelid Factor", 3f, 0f, 5f, (val) => {
                eyelidControl.lookUpTopEyelidFactor = val;
            }, true, "F2", false);

            SetUpMorphChooser("LeftTopEyelidUpMorphName", "Left Top Up Morph", eyelidControl.LeftTopEyelidUpMorphName,
                (string morphName) => {
                    if(FindMorph(morphName) == null) {
                        eyelidControl.LeftTopEyelidUpMorphName = morphName;
                    }
                }
            );
            SetUpMorphChooser("RightTopEyelidUpMorphName", "Right Top Up Morph", eyelidControl.RightTopEyelidUpMorphName,
                (string morphName) => {
                    if(FindMorph(morphName) == null) {
                        eyelidControl.RightTopEyelidUpMorphName = morphName;
                    }
                }
            );


            AddSlider("Look Up Bottom Eyelid Factor", 1f, 0f, 5f, (val) => {
                eyelidControl.lookUpBottomEyelidFactor = val;
            }, true, "F2", false);
            SetUpMorphChooser("LeftBottomEyelidUpMorphName", "Left Bottom Up Morph", eyelidControl.LeftBottomEyelidUpMorphName,
                (string morphName) => {
                    if(FindMorph(morphName) == null) {
                        eyelidControl.LeftBottomEyelidUpMorphName = morphName;
                    }
                }
            );
            SetUpMorphChooser("RightBottomEyelidUpMorphName", "Right Bottom Up Morph", eyelidControl.RightBottomEyelidUpMorphName,
                (string morphName) => {
                    if(FindMorph(morphName) == null) {
                        eyelidControl.RightBottomEyelidUpMorphName = morphName;
                    }
                }
            );



            AddSlider("Look Down Top Eyelid Factor", 1.5f, 0f, 5f, (val) => {
                eyelidControl.lookDownTopEyelidFactor = val;
            }, true, "F2", false);
            SetUpMorphChooser("LeftTopEyelidDownMorphName", "Left Top Down Morph", eyelidControl.LeftTopEyelidDownMorphName,
                (string morphName) => {
                    if(FindMorph(morphName) == null) {
                        eyelidControl.LeftTopEyelidDownMorphName = morphName;
                    }
                }
            );
            SetUpMorphChooser("RightTopEyelidDownMorphName", "Right Top Down Morph", eyelidControl.RightTopEyelidDownMorphName,
                (string morphName) => {
                    if(FindMorph(morphName) == null) {
                        eyelidControl.RightTopEyelidDownMorphName = morphName;
                    }
                }
            );



            AddSlider("Look Down Bottom Eyelid Factor", 4f, 0f, 5f, (val) => {
                eyelidControl.lookDownBottomEyelidFactor = val;
            }, true, "F2", false);
            SetUpMorphChooser("LeftBottomEyelidDownMorphName", "Left Bottom Down Morph", eyelidControl.LeftBottomEyelidDownMorphName,
                (string morphName) => {
                    if(FindMorph(morphName) == null) {
                        eyelidControl.LeftBottomEyelidDownMorphName = morphName;
                    }
                }
            );
            SetUpMorphChooser("RightBottomEyelidDownMorphName", "Right Bottom Down Morph", eyelidControl.RightBottomEyelidDownMorphName,
                (string morphName) => {
                    if(FindMorph(morphName) == null) {
                        eyelidControl.RightBottomEyelidDownMorphName = morphName;
                    }
                }
            );


            // Left and Right
            AddLabel("LeftRightControls", "Left And Right", true);
            AddSlider("Top Eyelid Look Inward Factor", 1f, 0f, 5f, (val) => {
                lookInwardTopEyelidFactor = val;
            }, true, "F2", true);
             SetUpMorphChooser("LeftTopEyelidRightMorphName", "Left Top Lid Look Inward", "None",
                (string morphName) => {
                    LeftTopEyelidRightMorph = FindMorph(morphName);
                }
            , true);
            SetUpMorphChooser("RightTopEyelidLeftMorphName", "Right Top Lid Look Inward", "None",
                (string morphName) => {
                    RightTopEyelidLeftMorph = FindMorph(morphName);
                }
            , true);
           

            AddSlider("Bottom Eyelid Look Inward Factor", 1f, 0f, 5f, (val) => {
                lookInwardBottomEyelidFactor = val;
            }, true, "F2", true);
            SetUpMorphChooser("LeftBottomEyelidRightMorphName", "Left Bottom Lid Look Inward", "None",
                (string morphName) => {
                    LeftBottomEyelidRightMorph = FindMorph(morphName);
                }
            , true);
            SetUpMorphChooser("RightBottomEyelidLeftMorph", "Right Bottom Lid Look Inward", "None",
                (string morphName) => {
                    RightBottomEyelidLeftMorph = FindMorph(morphName);
                }
            , true);



            AddSlider("Top Eyelid Look Outward Factor", 1f, 0f, 5f, (val) => {
                lookOutwardTopEyelidFactor = val;
            }, true, "F2", true);
            SetUpMorphChooser("LeftTopEyelidLeftMorphName", "Left Top Lid Look Outward", "None",
                (string morphName) => {
                    LeftTopEyelidLeftMorph = FindMorph(morphName);
                }
            , true);
            SetUpMorphChooser("RightTopEyelidRightMorphName", "Right Top Lid Look Outward", "None",
                (string morphName) => {
                    RightTopEyelidRightMorph = FindMorph(morphName);
                }
            , true);


            AddSlider("Bottom Eyelid Look Outward  Factor", 1f, 0f, 5f, (val) => {
                lookOutwardBottomEyelidFactor = val;
            }, true, "F2", true);
            SetUpMorphChooser("LeftBottomEyelidLeftMorph", "Left Bottom Lid Look Outward", "None",
                (string morphName) => {
                    LeftBottomEyelidLeftMorph = FindMorph(morphName);
                }
            , true);
            SetUpMorphChooser("RightBottomEyelidRightMorphName", "Right Bottom Lid Outward", "None",
                (string morphName) => {
                    RightBottomEyelidRightMorph = FindMorph(morphName);
                }
            , true);

             AddLabel("Spacer", "", true);
		}

        private void Update() {
            if(needRebuildMorphs) {
                RebuildMorphs();
                needRebuildMorphs = false;        
            }

            if(morphsMeedUpdate) {
                UpdateMorphs();
                morphsMeedUpdate = false;
            }

            // Left eye
            if (eyelidControl.leftEye != null && eyelidControl.GetBoolJSONParam("eyelidLookMorphsEnabled").val) {
                float y = Quaternion2Angles.GetAngles(eyelidControl.leftEye.localRotation, Quaternion2Angles.RotationOrder.ZYX).y;
                // Look left
                if (y < 0f) {
                    // Set look-left morphs 
                    if(LeftTopEyelidLeftMorph != null) {
                        LeftTopEyelidLeftMorph.morphValue = -y * lookOutwardTopEyelidFactor;
                    }
                    if(LeftBottomEyelidLeftMorph != null) {
                        LeftBottomEyelidLeftMorph.morphValue = -y * lookOutwardBottomEyelidFactor;
                    }

                    // Unset look-right morphs
                    if(LeftTopEyelidRightMorph != null) {
                        LeftTopEyelidRightMorph.morphValue = 0;
                    }

                    if(LeftBottomEyelidRightMorph != null) {
                        LeftBottomEyelidRightMorph.morphValue = 0;
                    }
                }
                // Look right
                else {
                    // Set look-right morphs 
                    if(LeftTopEyelidRightMorph != null) {
                        LeftTopEyelidRightMorph.morphValue = y * lookInwardTopEyelidFactor;
                    }

                    if(LeftBottomEyelidRightMorph != null) {
                        LeftBottomEyelidRightMorph.morphValue = y * lookInwardBottomEyelidFactor;
                    }

                    // Unset look-left morphs
                    if(LeftTopEyelidLeftMorph != null) {
                        LeftTopEyelidLeftMorph.morphValue = 0;
                    }
                    if(LeftBottomEyelidLeftMorph != null) {
                        LeftBottomEyelidLeftMorph.morphValue = 0;
                    }
                }
            }

            // Right eye
            if (eyelidControl.rightEye != null && eyelidControl.GetBoolJSONParam("eyelidLookMorphsEnabled").val) {
                float y = Quaternion2Angles.GetAngles(eyelidControl.rightEye.localRotation, Quaternion2Angles.RotationOrder.ZYX).y;
                // Look left
                if (y < 0f) {
                    // Set look-left morphs 
                    if(RightTopEyelidLeftMorph != null) {
                        RightTopEyelidLeftMorph.morphValue = -y * lookInwardTopEyelidFactor;
                    }
                    if(RightBottomEyelidLeftMorph != null) {
                        RightBottomEyelidLeftMorph.morphValue = -y * lookInwardBottomEyelidFactor;
                    }

                    // Unset look-right morphs
                    if(RightTopEyelidRightMorph != null) {
                        RightTopEyelidRightMorph.morphValue = 0;
                    }

                    if(RightBottomEyelidRightMorph != null) {
                        RightBottomEyelidRightMorph.morphValue = 0;
                    }
                }
                // Look right
                else {
                    // Set look-right morphs 
                    if(RightTopEyelidRightMorph != null) {
                        RightTopEyelidRightMorph.morphValue = y * lookOutwardTopEyelidFactor;
                    }

                    if(RightBottomEyelidRightMorph != null) {
                        RightBottomEyelidRightMorph.morphValue = y * lookOutwardBottomEyelidFactor;
                    }

                    // Unset look-left morphs
                    if(RightTopEyelidLeftMorph != null) {
                        RightTopEyelidLeftMorph.morphValue = 0;
                    }
                    if(RightBottomEyelidLeftMorph != null) {
                        RightBottomEyelidLeftMorph.morphValue = 0;
                    }
                }
            }

            /*
        private DAZMorph LeftTopEyelidLeftMorph;
        private DAZMorph RightTopEyelidLeftMorph;
        private DAZMorph LeftBottomEyelidLeftMorph;
        private DAZMorph RightBottomEyelidLeftMorph;
        private DAZMorph LeftTopEyelidRightMorph;
        private DAZMorph RightTopEyelidRightMorph;
        private DAZMorph LeftBottomEyelidRightMorph;
        private DAZMorph RightBottomEyelidRightMorph;
            */
        }

        private void SetUpMorphChooser(string storeName, string displayName, string defaultVal, JSONStorableStringChooser.SetStringCallback callback, bool rightside = false) {
            var chooser = new JSONStorableStringChooser(storeName, morphNames, defaultVal, displayName, callback);
		
			RegisterStringChooser(chooser);
			var popup = CreateFilterablePopup(chooser, rightside);
			popup.popup.onOpenPopupHandlers += () => {
                RefreshMorphList();
				chooser.choices = morphNames;
			};
        }

        private DAZMorph FindMorph(string morphName) {
            if(morphName == "None") {
                return dummyMorph;
            }

            var morph = eyelidControl.morphBank.GetMorph(morphName);
            if(morph == null) {
                SuperController.LogError($"Could not find morph {morphName}");
                return null;
            }

            if(morph.isRuntime) {
                morph.isRuntime = false;
                if(!morph.deltasLoaded) {
                    morph.LoadDeltas();
                }
                modifiedMorphs.Add(morph);
                needRebuildMorphs = true;
            }
            
            morphsMeedUpdate = true;
            return morph;
        }

        private void AddLabel(string name, string text, bool rightSide = false) {
            var textField = CreateTextField(new JSONStorableString(name, text), rightSide);
		    textField.height = 1;	
        }

        private void RebuildMorphs() {
            eyelidControl.morphBank.RebuildAllLookups();
        }
        private void UpdateMorphs() {
            var morphBank = eyelidControl.morphBank;
            eyelidControl.morphBank = null;
            eyelidControl.morphBank = morphBank;
        }

        private JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = false, string valueFormat = "F2", bool rightSide = false){
            JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
            storableFloat.storeType = JSONStorableParam.StoreType.Full;
            UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
            uiSlider.rangeAdjustEnabled = rangeAdjust;
            uiSlider.valueFormat = valueFormat;
            RegisterFloat(storableFloat);
            
            if(onChange != null){
                storableFloat.setCallbackFunction = onChange;
            }
            
            return storableFloat;
        }

        private void RefreshMorphList(){
            morphNames = new List<string>{"None"};
            morphNames.AddRange(eyelidControl.morphBank.morphs.Select((m) => m.morphName).ToList());
        }
		
		
		private void OnDisable()
		{
            foreach(var entry in floatParams) {
                var storableFloat = entry.Value;
                storableFloat.setCallbackFunction(storableFloat.defaultVal);
            }

            foreach(var entry in stringChooserParams) {
                var stringChooser = entry.Value;
                stringChooser.setCallbackFunction(stringChooser.defaultVal);
            }
            UpdateMorphs();
		}

        private void OnEnable()
		{
            foreach(var entry in floatParams) {
                var storableFloat = entry.Value;
                storableFloat.setCallbackFunction(storableFloat.val);
            }

            foreach(var entry in stringChooserParams) {
                var stringChooser = entry.Value;
                stringChooser.setCallbackFunction(stringChooser.val);
            }
		}

        private void OnDestroy(){
            foreach(var morph in modifiedMorphs) {
                morph.isRuntime = true;
            }
            RebuildMorphs();
        }

        private void LoadPreset(string path) {
			if (string.IsNullOrEmpty(path))
				return;
			
			var presetJson = (JSONClass)LoadJSON(path);
			
			RestoreFromJSON(presetJson, true, true, null, false);
			
			
		}
		
		private void SavePreset(string path) {
			SuperController.singleton.fileBrowserUI.fileFormat = null;
			if (string.IsNullOrEmpty(path)) {
				return;
			}

			if (!path.ToLower().EndsWith($".{fileExt}")) {
				path += $".{fileExt}";
			}

			SaveJSON(GetJSON(true,true,true), path);
		}
	}
}